//   RW Sysutils
// ================
// ______
// -The RW Foundational Libraries

// Filename: ./src/rw/util/utility.cpp
// -Definitions of some baseline RW Sysutils features

#include "incls_all.hpp"

namespace rw::util {

//------------------------------------------------------------------------------
void rw_error(std::string const& s) {
  if (use_stamps)
    analyze_callgraph();

  ostringstream ss{};
  ss << ">>>  ERROR  <<<\n - " << s << '\n';

  throw runtime_error(ss.str());
}

//------------------------------------------------------------------------------
void rw_error(std::string const& err_msg,
              Hip const&         tag,
              std::string const& func_nm) {
  if (use_stamps)
    analyze_callgraph();

  ostringstream ss{};
  ss << ">>>  ERROR  <<<\n - " << err_msg << "  [from : " << Hdl_id{tag}
     << ",  via : " << func_nm << "]\n";

  throw runtime_error(ss.str());
}

//------------------------------------------------------------------------------
void only_master_can(Hip const&         master,
                     Hip const&         stranger,
                     Hip const&         slave,
                     std::string const& func_nm,
                     std::string const& complaint) {
  puts("");
  prt_tag(master, "Master");
  prt_tag(stranger, "Stranger");
  rw_error(complaint, slave, func_nm);
}

//------------------------------------------------------------------------------
void analyze_callgraph() {  //

  ///
  // load Callgraph:

  Callgraph cgraph{};

  for (auto const prw : RW_objs) {  // get refs to all the active RW objects

    if (prw) {  // is the pointer still valid?
      auto const rw_obj{reference_to<RW>(prw)};
      cgraph.emplace_back(RW_stamps{rw_obj.t(), rw_obj.get_stamps()});
    }
  }

  ///
  // print Callgraph's items to console:

  cout << "Callgraph size (ie, the number of active RW objects) : "
       << cgraph.size() << "\n\n";

  for (auto& [out_tag, stamps] : cgraph) {  //
    Hdl_id const out_hid{out_tag};

    cout << " - Object's Handle.ID : '" << out_hid << "'\n";
    cout << "\tobject's stampstack size : " << stamps.size() << "\n";

    while (! stamps.empty()) {  //

      auto const& [in_tag, func_nm, tim]{stamps.top()};
      Hdl_id const in_hid{in_tag};

      cout << setw(24) << setfill(' ') << in_hid << '\t'  //
           << setw(6) << setfill(' ') << ((in_hid == out_hid) ? "(self)" : "")
           << "  calls obj's :"                    //
           << setw(22) << setfill(' ') << func_nm  //
           << " func at :  " << tim << "  utc\n";

      stamps.pop();
    }

    puts("");
  }
}

//------------------------------------------------------------------------------
void prt_tag(Hip const& tag, std::string const& msg) {
  Hdl_id const hid{tag};

  auto const f{cout.flags()};  // store flags state

  cout << setw(32) << msg << " - " << setw(24) << dec << hid
       << "    \taddr:  " << setw(16) << hex << tag.ptr << '\n';

  cout.flags(f);  // restore flags state
}

//------------------------------------------------------------------------------
void to_lower(std::string& in_out) {
  std::transform(
      cbegin(in_out), cend(in_out), begin(in_out),
      [](unsigned char c) -> unsigned char { return std::tolower(c); });
}

//------------------------------------------------------------------------------
auto is_empty(std::basic_stringstream<char>& ss) -> bool {  //

  auto posn = ss.tellp();  // store init posn
  //
  // DESIGN: check for failure here ??
  ss.seekp(0, std::ios_base::end);  // seek to stream's end
  auto chk = (ss.tellp() == 0);     // check if size == 0 ?
  //
  ss.seekp(posn);  // restore init posn

  return chk;
}

//------------------------------------------------------------------------------
auto is_empty(std::basic_fstream<char>& file) -> bool {
  return (file.peek() == ifstream::traits_type::eof());
}

//------------------------------------------------------------------------------
auto strmlines_2_vec(std::basic_stringstream<char>& strm)
    -> std::vector<std::string> {  //

  vector<string> res;
  //
  for (string line; getline(strm, line);)
    res.push_back(line);

  return res;
}

// DESIGN: address the code-envy here?

//------------------------------------------------------------------------------
auto strmlines_2_vec(std::basic_fstream<char>& strm)
    -> std::vector<std::string> {  //

  vector<string> res;
  //
  for (string line; getline(strm, line);)
    res.push_back(line);

  return res;
}

//------------------------------------------------------------------------------
auto operator<<(std::ostream& os, Time const& tim) -> std::ostream& {  //

  // just go through the duration and 'whittle-down' each part as an output var
  auto time_slice = tim.time_since_epoch();

  // until we can have C++20 on SBCs and other smol devices we'll just use this:
  using Days      = duration<long, ratio<86400, 1>>;
  auto const days = duration_cast<Days>(time_slice);
  time_slice -= days;
  ///////

  auto const hrs = duration_cast<hours>(time_slice);
  time_slice -= hrs;
  auto const mins = duration_cast<minutes>(time_slice);
  time_slice -= mins;
  auto const secs = duration_cast<seconds>(time_slice);
  time_slice -= secs;
  auto const millis = duration_cast<milliseconds>(time_slice);
  time_slice -= millis;
  auto const micros = duration_cast<microseconds>(time_slice);
  time_slice -= micros;
  auto const nanos = duration_cast<nanoseconds>(time_slice);

  return os << setw(2) << setfill('_') << hrs.count() << "h."      //
            << setw(2) << setfill('_') << mins.count() << "m."     //
            << setw(2) << setfill('_') << secs.count() << "s."     //
            << setw(3) << setfill('_') << millis.count() << "ms."  //
            << setw(3) << setfill('_') << micros.count() << "us."  //
            << setw(3) << setfill('_') << nanos.count() << "ns";

  // https://en.cppreference.com/w/cpp/chrono/duration
  // https://en.cppreference.com/w/cpp/chrono/duration/duration_cast
  // https://en.cppreference.com/w/cpp/numeric/ratio/ratio
}

//------------------------------------------------------------------------------
auto operator<<(std::stringstream& ss, std::vector<std::string> const& v)
    -> std::stringstream& {  //

  for (const auto& line : v)
    ss << line << '\n';

  return ss;
}

//------------------------------------------------------------------------------
auto operator<<(std::fstream& fs, std::vector<std::string> const& v)
    -> std::fstream& {  //

  for (const auto& line : v)
    fs << line << '\n';

  return fs;
}

//------------------------------------------------------------------------------
auto operator>>(std::stringstream& ss, std::vector<std::string>& v)
    -> std::stringstream& {  //

  for (string line; getline(ss, line);)
    v.push_back(line);

  return ss;
}

//------------------------------------------------------------------------------
auto operator>>(std::fstream& fs, std::vector<std::string>& v)
    -> std::fstream& {  //

  for (string line; getline(fs, line);)
    v.push_back(line);

  return fs;
}

//------------------------------------------------------------------------------
void trim_l(std::string& in_out, const char c) {
  auto it{begin(in_out)};
  if (*it == c)
    in_out.erase(it);
}

//------------------------------------------------------------------------------
void trim_r(std::string& in_out, const char c) {
  auto it{end(in_out) - 1};
  if (*it == c)
    in_out.erase(it);
}

//------------------------------------------------------------------------------
void trim_lr(std::string& in_out, const char c) {
  trim_l(in_out, c);
  trim_r(in_out, c);
}

}  // namespace rw::util

//    --=^=--

// "So is it true Sempai? Are robowaifus really better than any human woman?"

// rw_foundations_lib.v220212
// Copyright (2022)
// License MIT(Expat) https://opensource.org/licenses/MIT
